"""Companion APIs to create and convert integrity-checkable stores."""

from __future__ import annotations

import inspect
import shutil
from collections.abc import Iterable, Mapping
from pathlib import Path
from typing import Any, Literal
from uuid import uuid4

import numpy as np
import xarray as xr
import zarr
from zarr.core.config import config

from .integrity import check_store
from .layout import chunk_key, chunk_path, coord_in_bounds, expected_chunk_coords, scan_array_specs
from .manifest import MANIFEST_ROOT, load_variable_manifest, write_variable_manifest
from .models import ChunkRef, ConvertDirection, ConvertReport, CreateReport, NoDataStrategy


def _normalize_chunks(
    chunks: Mapping[str, Iterable[Iterable[int]]] | None,
) -> dict[str, list[tuple[int, ...]]]:
    normalized: dict[str, list[tuple[int, ...]]] = {}
    if not chunks:
        return normalized
    for variable, coords in chunks.items():
        parsed = [tuple(int(value) for value in coord) for coord in coords]
        normalized[str(variable)] = sorted(set(parsed))
    return normalized


def _to_zarr(
    dataset: xr.Dataset,
    store_path: Path,
    *,
    mode: str = "w",
    write_empty_chunks: bool = True,
    consolidated: bool = False,
    extra_kwargs: Mapping[str, Any] | None = None,
) -> None:
    params = inspect.signature(dataset.to_zarr).parameters
    kwargs: dict[str, Any] = {"store": store_path, "mode": mode}
    user_has_zarr_param = bool(
        extra_kwargs
        and ("zarr_format" in extra_kwargs or "zarr_version" in extra_kwargs)
    )
    if not user_has_zarr_param:
        if "zarr_format" in params:
            kwargs["zarr_format"] = 3
        elif "zarr_version" in params:
            kwargs["zarr_version"] = 3
        else:
            raise ValueError("xarray.to_zarr does not expose zarr_format/zarr_version")
    elif "zarr_format" not in params and "zarr_version" not in params:
        raise ValueError("xarray.to_zarr does not expose zarr_format/zarr_version")
    if "write_empty_chunks" in params and "write_empty_chunks" not in kwargs:
        kwargs["write_empty_chunks"] = write_empty_chunks
    if "consolidated" in params and "consolidated" not in kwargs:
        kwargs["consolidated"] = consolidated
    if extra_kwargs:
        for key, value in extra_kwargs.items():
            if key == "store":
                raise ValueError("'store' must not be provided in extra to_zarr kwargs")
            kwargs[key] = value
    dataset.to_zarr(**kwargs)


def _write_dataset(dataset: xr.Dataset, store_path: Path) -> None:
    _to_zarr(dataset, store_path)


def _delete_chunk_file(chunk_file: Path, array_root: Path) -> None:
    if not chunk_file.exists():
        return
    chunk_file.unlink()
    current = chunk_file.parent
    while current != array_root and current.exists() and not any(current.iterdir()):
        current.rmdir()
        current = current.parent


def _staging_manifest_root() -> Path:
    return MANIFEST_ROOT.parent / f"manifests.staging-{uuid4().hex}"


def _backup_manifest_root() -> Path:
    return MANIFEST_ROOT.parent / f"manifests.backup-{uuid4().hex}"


def _prepare_staged_manifests(store: Path, staged_root: Path) -> Path:
    staged_dir = store / staged_root
    source_dir = store / MANIFEST_ROOT
    if source_dir.exists():
        shutil.copytree(source_dir, staged_dir)
    else:
        staged_dir.mkdir(parents=True, exist_ok=True)
    return staged_dir


def _swap_staged_manifests(store: Path, staged_root: Path) -> None:
    active_dir = store / MANIFEST_ROOT
    staged_dir = store / staged_root
    backup_root = _backup_manifest_root()
    backup_dir = store / backup_root

    moved_active = False
    try:
        if active_dir.exists():
            active_dir.replace(backup_dir)
            moved_active = True
        staged_dir.replace(active_dir)
    except Exception:
        if moved_active and backup_dir.exists() and not active_dir.exists():
            backup_dir.replace(active_dir)
        if staged_dir.exists():
            shutil.rmtree(staged_dir, ignore_errors=True)
        raise
    finally:
        if backup_dir.exists():
            shutil.rmtree(backup_dir, ignore_errors=True)


def _update_store_metadata_in_place(
    store: Path,
    *,
    no_data_chunks: dict[str, list[tuple[int, ...]]],
    infer_no_data_from_store: bool = False,
) -> CreateReport:
    if infer_no_data_from_store and no_data_chunks:
        raise ValueError(
            "infer_no_data_from_store=True cannot be combined with explicit no_data_chunks"
        )

    specs = {spec.name: spec for spec in scan_array_specs(store)}
    if not infer_no_data_from_store:
        unknown = sorted(set(no_data_chunks) - set(specs))
        if unknown:
            raise ValueError(f"Unknown variables in no_data_chunks: {', '.join(unknown)}")

    report = CreateReport(store_path=str(store), no_data_strategy="manifest")
    staged_root = _staging_manifest_root()
    if infer_no_data_from_store:
        staged_dir = store / staged_root
        staged_dir.mkdir(parents=True, exist_ok=True)
    else:
        staged_dir = _prepare_staged_manifests(store, staged_root)

    try:
        if infer_no_data_from_store:
            for spec in specs.values():
                refs: list[ChunkRef] = []
                for coord in expected_chunk_coords(spec):
                    if chunk_path(spec, coord).exists():
                        continue
                    refs.append(ChunkRef(coord=coord, key=chunk_key(spec, coord)))
                if not refs:
                    continue
                staged_path = write_variable_manifest(
                    store,
                    spec.name,
                    refs,
                    manifest_root=staged_root,
                )
                report.manifests_written.append(str(store / MANIFEST_ROOT / staged_path.name))
        else:
            for variable, coords in no_data_chunks.items():
                spec = specs[variable]
                _, existing_refs = load_variable_manifest(
                    store,
                    variable,
                    manifest_root=staged_root,
                )
                merged = {ref.coord: ref for ref in existing_refs}

                for coord in coords:
                    if not coord_in_bounds(spec, coord):
                        raise ValueError(
                            f"Chunk coord {coord} out of bounds for variable {variable}"
                        )
                    ref = ChunkRef(coord=coord, key=chunk_key(spec, coord))
                    if chunk_path(spec, coord).exists():
                        raise ValueError(
                            "In-place manifest update only accepts currently missing chunks: "
                            f"{variable}:{coord}"
                        )
                    merged[coord] = ref

                merged_refs = [merged[coord] for coord in sorted(merged)]
                staged_path = write_variable_manifest(
                    store,
                    variable,
                    merged_refs,
                    manifest_root=staged_root,
                )
                report.manifests_written.append(str(store / MANIFEST_ROOT / staged_path.name))

        integrity = check_store(store, _manifest_root=staged_root)
        if not integrity.ok:
            raise RuntimeError("In-place metadata update would fail integrity validation")

        _swap_staged_manifests(store, staged_root)
        return report
    except Exception:
        if staged_dir.exists():
            shutil.rmtree(staged_dir, ignore_errors=True)
        raise


def create_store(
    dataset: xr.Dataset | None,
    store_path: str | Path,
    *,
    no_data_chunks: Mapping[str, Iterable[Iterable[int]]] | None = None,
    no_data_strategy: NoDataStrategy = "manifest",
    overwrite: bool = False,
    in_place_metadata_only: bool = False,
    infer_no_data_from_store: bool = False,
) -> CreateReport:
    """Create a Zarr v3 store with explicit no-data policy."""

    if no_data_strategy not in {"manifest", "empty_chunks"}:
        raise ValueError("no_data_strategy must be 'manifest' or 'empty_chunks'")

    normalized = _normalize_chunks(no_data_chunks)
    store = Path(store_path)

    if in_place_metadata_only:
        if no_data_strategy != "manifest":
            raise ValueError("in_place_metadata_only requires no_data_strategy='manifest'")
        if overwrite:
            raise ValueError("in_place_metadata_only does not support overwrite")
        if not store.exists():
            raise FileNotFoundError(f"Store does not exist: {store}")
        if not store.is_dir():
            raise NotADirectoryError(f"Store path is not a directory: {store}")
        return _update_store_metadata_in_place(
            store,
            no_data_chunks=normalized,
            infer_no_data_from_store=infer_no_data_from_store,
        )

    if infer_no_data_from_store:
        raise ValueError(
            "infer_no_data_from_store is only supported with in_place_metadata_only=True"
        )

    if dataset is None:
        raise ValueError("dataset is required unless in_place_metadata_only=True")

    if store.exists():
        if not overwrite:
            raise FileExistsError(f"Store already exists: {store}")
        shutil.rmtree(store)

    store.parent.mkdir(parents=True, exist_ok=True)
    _write_dataset(dataset, store)

    specs = {spec.name: spec for spec in scan_array_specs(store)}
    unknown = sorted(set(normalized) - set(specs))
    if unknown:
        raise ValueError(f"Unknown variables in no_data_chunks: {', '.join(unknown)}")

    report = CreateReport(store_path=str(store), no_data_strategy=no_data_strategy)

    for variable, coords in normalized.items():
        spec = specs[variable]
        refs: list[ChunkRef] = []
        removed: list[ChunkRef] = []
        for coord in coords:
            if not coord_in_bounds(spec, coord):
                raise ValueError(f"Chunk coord {coord} out of bounds for variable {variable}")
            ref = ChunkRef(coord=coord, key=chunk_key(spec, coord))
            refs.append(ref)
            if no_data_strategy == "manifest":
                _delete_chunk_file(chunk_path(spec, coord), spec.path)
                removed.append(ref)
            elif not chunk_path(spec, coord).exists():
                raise RuntimeError(
                    "Expected chunk file missing after write_empty_chunks=True: "
                    f"{variable}:{coord}"
                )

        if no_data_strategy == "manifest":
            manifest_file = write_variable_manifest(store, variable, refs)
            report.manifests_written.append(str(manifest_file))
            report.removed_chunks[variable] = removed

    integrity = check_store(store)
    if not integrity.ok:
        raise RuntimeError("Created store failed integrity validation")

    return report


def guarded_to_zarr(
    dataset: xr.Dataset,
    store_path: str | Path,
    *,
    overwrite: bool = False,
    infer_no_data_from_store: bool = True,
    no_data_chunks: Mapping[str, Iterable[Iterable[int]]] | None = None,
    to_zarr_kwargs: Mapping[str, Any] | None = None,
) -> CreateReport:
    """Write a Zarr v3 store and immediately guard it with xzarrguard manifests.

    By default, this treats the freshly written store as ground truth and records
    all currently missing chunks as allowed-missing manifests.
    """

    store = Path(store_path)
    if store.exists():
        if not overwrite:
            raise FileExistsError(f"Store already exists: {store}")
        shutil.rmtree(store)

    store.parent.mkdir(parents=True, exist_ok=True)
    _to_zarr(dataset, store, extra_kwargs=to_zarr_kwargs)
    return create_store(
        None,
        store,
        no_data_chunks=no_data_chunks,
        in_place_metadata_only=True,
        infer_no_data_from_store=infer_no_data_from_store,
    )


def _manifest_files(store: Path) -> list[Path]:
    manifest_dir = store / MANIFEST_ROOT
    if not manifest_dir.exists():
        return []
    return sorted(path for path in manifest_dir.glob("*.json") if path.is_file())


def _detect_convert_direction(store: Path) -> ConvertDirection:
    if _manifest_files(store):
        return "manifest_to_materialized"
    return "materialized_to_manifest"


def _chunk_slices(spec: Any, coord: tuple[int, ...]) -> tuple[slice, ...]:
    return tuple(
        slice(index * chunk, min((index + 1) * chunk, size))
        for index, chunk, size in zip(coord, spec.chunk_shape, spec.shape, strict=True)
    )


def _open_array_node(group: Any, array_name: str) -> Any:
    node = group
    for part in array_name.split("/") if array_name else []:
        node = node[part]
    return node


def _dtype_supports_nan(dtype: np.dtype[Any]) -> bool:
    return bool(np.issubdtype(dtype, np.floating) or np.issubdtype(dtype, np.complexfloating))


def _collect_all_nan_chunks(store: Path) -> dict[str, list[ChunkRef]]:
    specs = {spec.name: spec for spec in scan_array_specs(store)}
    root = zarr.open_group(str(store), mode="r")

    result: dict[str, list[ChunkRef]] = {}
    for spec in specs.values():
        array = _open_array_node(root, spec.name)
        dtype = np.dtype(array.dtype)
        if not _dtype_supports_nan(dtype):
            continue

        refs: list[ChunkRef] = []
        for coord in expected_chunk_coords(spec):
            if not chunk_path(spec, coord).exists():
                continue
            data = np.asarray(array[_chunk_slices(spec, coord)])
            if data.size == 0:
                continue
            if np.isnan(data).all():
                refs.append(ChunkRef(coord=coord, key=chunk_key(spec, coord)))

        if refs:
            result[spec.name] = refs
    return result


def _fill_value_for_materialization(array: Any) -> Any:
    dtype = np.dtype(array.dtype)
    fill_value = getattr(array, "fill_value", None)
    if fill_value is not None:
        return fill_value
    if _dtype_supports_nan(dtype):
        return np.nan
    raise ValueError(
        f"Cannot materialize missing chunks for dtype {dtype}; no fill_value is available"
    )


def _convert_materialized_to_manifest(store: Path) -> ConvertReport:
    if _manifest_files(store):
        raise ValueError(
            "Store already has manifests; use direction='manifest_to_materialized' "
            "or direction='auto'"
        )

    integrity = check_store(store)
    if not integrity.ok:
        raise ValueError(
            "Store must pass integrity before materialized_to_manifest conversion"
        )

    specs = {spec.name: spec for spec in scan_array_specs(store)}
    deleted_chunks = _collect_all_nan_chunks(store)

    for variable, refs in deleted_chunks.items():
        spec = specs[variable]
        for ref in refs:
            _delete_chunk_file(chunk_path(spec, ref.coord), spec.path)

    create_report = create_store(
        None,
        store,
        in_place_metadata_only=True,
        infer_no_data_from_store=True,
    )

    return ConvertReport(
        store_path=str(store),
        direction="materialized_to_manifest",
        manifests_written=create_report.manifests_written,
        deleted_chunks=deleted_chunks,
    )


def _convert_manifest_to_materialized(store: Path) -> ConvertReport:
    manifest_files = _manifest_files(store)
    if not manifest_files:
        return ConvertReport(
            store_path=str(store),
            direction="manifest_to_materialized",
        )

    specs = {spec.name: spec for spec in scan_array_specs(store)}

    materialized_chunks: dict[str, list[ChunkRef]] = {}
    with config.set({"array.write_empty_chunks": True}):
        root = zarr.open_group(str(store), mode="r+")
        for spec in specs.values():
            has_manifest, refs = load_variable_manifest(store, spec.name)
            if not has_manifest:
                continue

            array = _open_array_node(root, spec.name)
            dtype = np.dtype(array.dtype)
            fill_value = _fill_value_for_materialization(array)

            materialized_refs: list[ChunkRef] = []
            for ref in refs:
                if chunk_path(spec, ref.coord).exists():
                    continue
                slices = _chunk_slices(spec, ref.coord)
                chunk_shape = tuple(item.stop - item.start for item in slices)
                if chunk_shape:
                    payload = np.full(chunk_shape, fill_value, dtype=dtype)
                else:
                    payload = np.asarray(fill_value, dtype=dtype)
                array[slices] = payload
                materialized_refs.append(ChunkRef(coord=ref.coord, key=chunk_key(spec, ref.coord)))

            if materialized_refs:
                materialized_chunks[spec.name] = materialized_refs

    manifest_dir = store / MANIFEST_ROOT
    if manifest_dir.exists():
        shutil.rmtree(manifest_dir)

    integrity = check_store(store, strict_stale_manifest=True)
    if not integrity.ok:
        raise RuntimeError("Converted store failed integrity validation")

    return ConvertReport(
        store_path=str(store),
        direction="manifest_to_materialized",
        manifests_removed=[str(path) for path in manifest_files],
        materialized_chunks=materialized_chunks,
    )


def convert_store(
    store_path: str | Path,
    *,
    direction: ConvertDirection | Literal["auto"] = "auto",
) -> ConvertReport:
    """Convert a store between materialized and manifest forms."""

    store = Path(store_path)
    if not store.exists():
        raise FileNotFoundError(f"Store does not exist: {store}")
    if not store.is_dir():
        raise NotADirectoryError(f"Store path is not a directory: {store}")

    if direction == "auto":
        selected = _detect_convert_direction(store)
    elif direction in {"materialized_to_manifest", "manifest_to_materialized"}:
        selected = direction
    else:
        raise ValueError(
            "direction must be 'auto', 'materialized_to_manifest', or 'manifest_to_materialized'"
        )

    if selected == "materialized_to_manifest":
        return _convert_materialized_to_manifest(store)
    return _convert_manifest_to_materialized(store)



from __future__ import annotations

import json
from pathlib import Path

import pytest

from xzarrguard.layout import scan_array_specs


def _write_json(path: Path, payload: dict) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(json.dumps(payload), encoding="utf-8")


def _group_meta(*, metadata: dict[str, dict] | None = None) -> dict:
    payload: dict[str, object] = {"zarr_format": 3, "node_type": "group", "attributes": {}}
    if metadata is not None:
        payload["consolidated_metadata"] = {
            "kind": "inline",
            "must_understand": False,
            "metadata": metadata,
        }
    return payload


def _array_meta(shape: tuple[int, ...], chunk_shape: tuple[int, ...]) -> dict:
    return {
        "zarr_format": 3,
        "node_type": "array",
        "shape": list(shape),
        "chunk_grid": {"name": "regular", "configuration": {"chunk_shape": list(chunk_shape)}},
        "chunk_key_encoding": {"name": "default", "configuration": {"separator": "/"}},
    }


def test_scan_array_specs_discovers_nested_arrays_without_scanning_chunks(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"
    _write_json(store / "zarr.json", _group_meta())
    _write_json(store / "a" / "zarr.json", _array_meta((4, 4), (2, 2)))
    _write_json(store / "group" / "zarr.json", _group_meta())
    _write_json(store / "group" / "b" / "zarr.json", _array_meta((8,), (4,)))

    # If chunk directories were recursively scanned, this would trigger a format error.
    _write_json(store / "a" / "c" / "0" / "zarr.json", {"zarr_format": 2, "node_type": "array"})

    specs = scan_array_specs(store)
    assert [item.name for item in specs] == ["a", "group/b"]


def test_scan_array_specs_rejects_non_v3_root(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"
    _write_json(store / "zarr.json", {"zarr_format": 2, "node_type": "group"})

    with pytest.raises(ValueError, match="zarr_format=3"):
        scan_array_specs(store)


def test_scan_array_specs_supports_consolidated_metadata_without_child_files(
    tmp_path: Path,
) -> None:
    store = tmp_path / "store.zarr"
    _write_json(
        store / "zarr.json",
        _group_meta(
            metadata={
                "a": _array_meta((4, 4), (2, 2)),
                "nested/b": _array_meta((8,), (4,)),
            }
        ),
    )

    specs = scan_array_specs(store)

    assert [item.name for item in specs] == ["a", "nested/b"]
    assert specs[0].path == store / "a"
    assert specs[1].path == store / "nested" / "b"


def test_scan_array_specs_prefers_consolidated_metadata_when_available(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"
    _write_json(
        store / "zarr.json",
        _group_meta(
            metadata={
                "a": _array_meta((4, 4), (2, 2)),
            }
        ),
    )
    # Broken per-array metadata should not matter if consolidated metadata is present.
    _write_json(store / "a" / "zarr.json", {"zarr_format": 2, "node_type": "array"})

    specs = scan_array_specs(store)

    assert [item.name for item in specs] == ["a"]



"""Public report models for xzarrguard."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Literal

type NoDataStrategy = Literal["manifest", "empty_chunks"]
type ConvertDirection = Literal["materialized_to_manifest", "manifest_to_materialized"]


@dataclass(frozen=True, slots=True)
class ChunkRef:
    """Reference to one logical chunk."""

    coord: tuple[int, ...]
    key: str

    def to_dict(self) -> dict[str, Any]:
        return {"coord": list(self.coord), "key": self.key}


@dataclass(slots=True)
class VariableIntegrity:
    """Per-variable integrity results."""

    name: str
    expected_chunks: int
    has_manifest: bool
    missing_unexpected: list[ChunkRef] = field(default_factory=list)
    missing_allowed: list[ChunkRef] = field(default_factory=list)
    stale_manifest: list[ChunkRef] = field(default_factory=list)
    manifest_key_mismatch: list[ChunkRef] = field(default_factory=list)
    manifest_out_of_bounds: list[ChunkRef] = field(default_factory=list)
    ok: bool = True

    def to_dict(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "ok": self.ok,
            "has_manifest": self.has_manifest,
            "expected_chunks": self.expected_chunks,
            "missing_unexpected": [item.to_dict() for item in self.missing_unexpected],
            "missing_allowed": [item.to_dict() for item in self.missing_allowed],
            "stale_manifest": [item.to_dict() for item in self.stale_manifest],
            "manifest_key_mismatch": [item.to_dict() for item in self.manifest_key_mismatch],
            "manifest_out_of_bounds": [item.to_dict() for item in self.manifest_out_of_bounds],
        }


@dataclass(slots=True)
class IntegrityReport:
    """Integrity report for a store."""

    store_path: str
    strict_stale_manifest: bool
    variables: dict[str, VariableIntegrity] = field(default_factory=dict)
    errors: list[str] = field(default_factory=list)
    timing: IntegrityTiming | None = None
    ok: bool = True

    def __bool__(self) -> bool:
        return self.ok

    def to_dict(self) -> dict[str, Any]:
        payload = {
            "store_path": self.store_path,
            "strict_stale_manifest": self.strict_stale_manifest,
            "ok": self.ok,
            "errors": list(self.errors),
            "variables": {name: report.to_dict() for name, report in self.variables.items()},
        }
        if self.timing is not None:
            payload["timing"] = self.timing.to_dict()
        return payload


@dataclass(slots=True)
class VariableTiming:
    """Timing details for one variable check."""

    manifest_load_s: float = 0.0
    manifest_validate_s: float = 0.0
    chunk_scan_s: float = 0.0
    expected_chunks: int = 0
    missing_chunks: int = 0

    def to_dict(self) -> dict[str, Any]:
        return {
            "manifest_load_s": self.manifest_load_s,
            "manifest_validate_s": self.manifest_validate_s,
            "chunk_scan_s": self.chunk_scan_s,
            "expected_chunks": self.expected_chunks,
            "missing_chunks": self.missing_chunks,
        }


@dataclass(slots=True)
class IntegrityTiming:
    """Optional coarse-grained timing information for `check_store`."""

    total_s: float = 0.0
    scan_specs_s: float = 0.0
    manifest_s: float = 0.0
    chunk_scan_s: float = 0.0
    exists_calls: int = 0
    variables: dict[str, VariableTiming] = field(default_factory=dict)

    def to_dict(self) -> dict[str, Any]:
        return {
            "total_s": self.total_s,
            "scan_specs_s": self.scan_specs_s,
            "manifest_s": self.manifest_s,
            "chunk_scan_s": self.chunk_scan_s,
            "exists_calls": self.exists_calls,
            "variables": {name: item.to_dict() for name, item in self.variables.items()},
        }


@dataclass(slots=True)
class CreateReport:
    """Result of creating a store with optional no-data policy."""

    store_path: str
    no_data_strategy: NoDataStrategy
    manifests_written: list[str] = field(default_factory=list)
    removed_chunks: dict[str, list[ChunkRef]] = field(default_factory=dict)
    ok: bool = True

    def to_dict(self) -> dict[str, Any]:
        return {
            "store_path": self.store_path,
            "no_data_strategy": self.no_data_strategy,
            "ok": self.ok,
            "manifests_written": list(self.manifests_written),
            "removed_chunks": {
                name: [item.to_dict() for item in refs]
                for name, refs in self.removed_chunks.items()
            },
        }


@dataclass(slots=True)
class ConvertReport:
    """Result of converting a store between materialized and manifest forms."""

    store_path: str
    direction: ConvertDirection
    manifests_written: list[str] = field(default_factory=list)
    manifests_removed: list[str] = field(default_factory=list)
    deleted_chunks: dict[str, list[ChunkRef]] = field(default_factory=dict)
    materialized_chunks: dict[str, list[ChunkRef]] = field(default_factory=dict)
    ok: bool = True

    def to_dict(self) -> dict[str, Any]:
        return {
            "store_path": self.store_path,
            "direction": self.direction,
            "ok": self.ok,
            "manifests_written": list(self.manifests_written),
            "manifests_removed": list(self.manifests_removed),
            "deleted_chunks": {
                name: [item.to_dict() for item in refs]
                for name, refs in self.deleted_chunks.items()
            },
            "materialized_chunks": {
                name: [item.to_dict() for item in refs]
                for name, refs in self.materialized_chunks.items()
            },
        }



"""xzarrguard public API."""

from ._version import __version__
from .create import convert_store, create_store, guarded_to_zarr
from .integrity import check_store
from .manifest import dump_no_data_chunks, load_no_data_chunks
from .models import (
    ChunkRef,
    ConvertDirection,
    ConvertReport,
    CreateReport,
    IntegrityReport,
    IntegrityTiming,
    NoDataStrategy,
    VariableIntegrity,
    VariableTiming,
)

__all__ = [
    "ChunkRef",
    "ConvertDirection",
    "ConvertReport",
    "CreateReport",
    "IntegrityReport",
    "IntegrityTiming",
    "NoDataStrategy",
    "VariableIntegrity",
    "VariableTiming",
    "__version__",
    "check_store",
    "convert_store",
    "create_store",
    "dump_no_data_chunks",
    "guarded_to_zarr",
    "load_no_data_chunks",
]



"""Remote-store helpers for integrity checks."""

from __future__ import annotations

import json
import posixpath
from collections.abc import Mapping
from pathlib import PurePosixPath
from typing import Any
from urllib.parse import quote

from fsspec.core import url_to_fs

from .layout import ArraySpec, _parse_array_spec
from .manifest import MANIFEST_SCHEMA_VERSION
from .models import ChunkRef


def is_remote_store_path(store_path: str | PurePosixPath | Any) -> bool:
    """Return whether the supplied store path is an fsspec-style URI."""

    return isinstance(store_path, str) and "://" in store_path


def open_remote_store(
    store_path: str,
    *,
    storage_options: Mapping[str, Any] | None = None,
) -> tuple[Any, str]:
    """Return the filesystem and root key for a remote store URI."""

    fs, root = url_to_fs(store_path, **dict(storage_options or {}))
    return fs, _normalize_key(root)


def scan_remote_array_specs(fs: Any, root_key: str) -> list[ArraySpec]:
    """Return every array spec found in a remote Zarr v3 store."""

    specs = _scan_remote_from_consolidated_metadata(fs, root_key)
    if specs:
        specs.sort(key=lambda item: item.name)
        return specs

    zarr_json_key = _join_key(root_key, "zarr.json")
    zarr_json_suffix = "/zarr.json"
    specs = []
    for meta_key in sorted(fs.find(root_key)):
        normalized_key = _normalize_key(meta_key)
        if normalized_key != zarr_json_key and not normalized_key.endswith(zarr_json_suffix):
            continue

        node_dir = _normalize_key(posixpath.dirname(normalized_key))
        rel_dir = posixpath.relpath(node_dir, root_key)
        if rel_dir != "." and ".xzarrguard" in PurePosixPath(rel_dir).parts:
            continue

        payload = json.loads(_read_text(fs, normalized_key))
        spec = _parse_array_spec(
            store_path=PurePosixPath(root_key),
            array_name="",
            array_path=PurePosixPath(node_dir),
            payload=payload,
            source=normalized_key,
        )
        if spec is not None:
            specs.append(spec)

    specs.sort(key=lambda item: item.name)
    return specs


def load_remote_variable_manifest(
    fs: Any,
    root_key: str,
    variable: str,
    *,
    manifest_root: str | PurePosixPath,
) -> tuple[bool, list[ChunkRef]]:
    """Read a single manifest from a remote store."""

    manifest_key = _join_key(
        root_key,
        _manifest_root_text(manifest_root),
        f"{quote(variable, safe='')}.json",
    )
    if not fs.exists(manifest_key):
        return False, []

    payload = json.loads(_read_text(fs, manifest_key))
    if payload.get("schema_version") != MANIFEST_SCHEMA_VERSION:
        raise ValueError(f"Unsupported manifest schema in {manifest_key}")
    items = payload.get("allowed_missing", [])
    refs = [
        ChunkRef(
            coord=tuple(int(value) for value in item["coord"]),
            key=str(item["key"]),
        )
        for item in items
    ]
    return True, refs


def remote_existing_chunk_keys(fs: Any, spec: ArraySpec) -> set[str]:
    """Return the chunk keys currently materialized for one remote array."""

    array_root = _spec_path_text(spec)
    zarr_json_key = _join_key(array_root, "zarr.json")
    result: set[str] = set()
    for key in fs.find(array_root):
        normalized = _normalize_key(key)
        if normalized == zarr_json_key:
            continue
        rel = posixpath.relpath(normalized, array_root)
        if rel in {".", "zarr.json"}:
            continue
        result.add(rel)
    return result


def _scan_remote_from_consolidated_metadata(fs: Any, root_key: str) -> list[ArraySpec]:
    root_meta_key = _join_key(root_key, "zarr.json")
    if not fs.exists(root_meta_key):
        return []

    root_payload = json.loads(_read_text(fs, root_meta_key))
    if root_payload.get("zarr_format") != 3:
        raise ValueError(f"Only zarr_format=3 is supported: {root_meta_key}")

    consolidated = root_payload.get("consolidated_metadata")
    if not isinstance(consolidated, dict):
        return []
    metadata = consolidated.get("metadata")
    if not isinstance(metadata, dict):
        return []

    specs: list[ArraySpec] = []
    for name, payload in metadata.items():
        if not isinstance(name, str) or not isinstance(payload, dict):
            continue
        array_path = PurePosixPath(_join_key(root_key, name)) if name else PurePosixPath(root_key)
        spec = _parse_array_spec(
            store_path=PurePosixPath(root_key),
            array_name=name,
            array_path=array_path,
            payload=payload,
            source=f"{root_meta_key} consolidated_metadata[{name!r}]",
        )
        if spec is not None:
            specs.append(spec)
    return specs


def _manifest_root_text(manifest_root: str | PurePosixPath) -> str:
    if isinstance(manifest_root, PurePosixPath):
        return manifest_root.as_posix().strip("/")
    return str(manifest_root).replace("\\", "/").strip("/")


def _normalize_key(key: str) -> str:
    if key in {"", "."}:
        return ""
    normalized = posixpath.normpath(key)
    if normalized == ".":
        return ""
    return normalized


def _join_key(base: str, *parts: str) -> str:
    current = _normalize_key(base)
    for part in parts:
        cleaned = str(part).replace("\\", "/").strip("/")
        if not cleaned:
            continue
        if current in {"", "."}:
            current = cleaned
        elif current == "/":
            current = f"/{cleaned}"
        else:
            current = posixpath.join(current, cleaned)
    return current


def _read_text(fs: Any, key: str) -> str:
    with fs.open(key, "rb") as handle:
        payload = handle.read()
    if isinstance(payload, bytes):
        return payload.decode("utf-8")
    return str(payload)


def _spec_path_text(spec: ArraySpec) -> str:
    path = spec.path
    if isinstance(path, PurePosixPath):
        return path.as_posix()
    return str(path).replace("\\", "/")



"""Zarr v3 local-store layout helpers."""

from __future__ import annotations

import json
import math
from dataclasses import dataclass
from itertools import product
from pathlib import Path
from typing import Any


@dataclass(frozen=True, slots=True)
class ArraySpec:
    """Minimal metadata needed for chunk validation."""

    name: str
    path: Path
    shape: tuple[int, ...]
    chunk_shape: tuple[int, ...]
    chunk_key_encoding: str
    separator: str


def _parse_array_spec(
    *,
    store_path: Path,
    array_name: str,
    array_path: Path,
    payload: dict[str, Any],
    source: Path | str,
) -> ArraySpec | None:
    if payload.get("zarr_format") != 3:
        raise ValueError(f"Only zarr_format=3 is supported: {source}")
    if payload.get("node_type") != "array":
        return None

    shape = tuple(int(v) for v in payload["shape"])
    chunk_grid = payload.get("chunk_grid", {})
    if chunk_grid.get("name") != "regular":
        raise ValueError(f"Only regular chunk grids are supported: {source}")
    chunk_shape = tuple(int(v) for v in chunk_grid["configuration"]["chunk_shape"])

    encoding = payload.get("chunk_key_encoding") or {
        "name": "default",
        "configuration": {"separator": "/"},
    }
    encoding_name = str(encoding.get("name", "default"))
    config = encoding.get("configuration", {})
    default_separator = "/" if encoding_name == "default" else "."
    separator = str(config.get("separator", default_separator))

    if not array_name:
        rel_dir = array_path.relative_to(store_path)
        array_name = "/".join(rel_dir.parts)

    return ArraySpec(
        name=array_name,
        path=array_path,
        shape=shape,
        chunk_shape=chunk_shape,
        chunk_key_encoding=encoding_name,
        separator=separator,
    )


def _scan_from_consolidated_metadata(store_path: Path) -> list[ArraySpec]:
    root_meta_path = store_path / "zarr.json"
    if not root_meta_path.exists():
        return []

    root_payload = json.loads(root_meta_path.read_text(encoding="utf-8"))
    if root_payload.get("zarr_format") != 3:
        raise ValueError(f"Only zarr_format=3 is supported: {root_meta_path}")

    consolidated = root_payload.get("consolidated_metadata")
    if not isinstance(consolidated, dict):
        return []
    metadata = consolidated.get("metadata")
    if not isinstance(metadata, dict):
        return []

    specs: list[ArraySpec] = []
    for name, payload in metadata.items():
        if not isinstance(name, str) or not isinstance(payload, dict):
            continue
        array_path = store_path / Path(name) if name else store_path
        spec = _parse_array_spec(
            store_path=store_path,
            array_name=name,
            array_path=array_path,
            payload=payload,
            source=f"{root_meta_path} consolidated_metadata[{name!r}]",
        )
        if spec is not None:
            specs.append(spec)
    return specs


def scan_array_specs(store_path: Path) -> list[ArraySpec]:
    """Return every array spec found in a local Zarr v3 store."""

    specs = _scan_from_consolidated_metadata(store_path)
    if specs:
        specs.sort(key=lambda item: item.name)
        return specs

    specs: list[ArraySpec] = []
    stack: list[Path] = [store_path]
    while stack:
        node_path = stack.pop()
        if ".xzarrguard" in node_path.parts:
            continue

        meta_path = node_path / "zarr.json"
        if not meta_path.exists():
            continue

        payload = json.loads(meta_path.read_text(encoding="utf-8"))
        node_type = payload.get("node_type")
        if node_type == "group":
            for child in sorted(node_path.iterdir(), reverse=True):
                if not child.is_dir() or child.name == ".xzarrguard":
                    continue
                child_meta = child / "zarr.json"
                if child_meta.exists():
                    stack.append(child)
            continue
        spec = _parse_array_spec(
            store_path=store_path,
            array_name="",
            array_path=meta_path.parent,
            payload=payload,
            source=meta_path,
        )
        if spec is not None:
            specs.append(spec)
    specs.sort(key=lambda item: item.name)
    return specs


def chunk_counts(spec: ArraySpec) -> tuple[int, ...]:
    """Return number of chunks per dimension."""

    if len(spec.shape) != len(spec.chunk_shape):
        raise ValueError(f"Shape/chunk rank mismatch for {spec.name}")
    return tuple(
        math.ceil(size / chunk) for size, chunk in zip(spec.shape, spec.chunk_shape, strict=True)
    )


def expected_chunk_coords(spec: ArraySpec):
    """Yield all expected chunk coordinates."""

    counts = chunk_counts(spec)
    if not counts:
        yield ()
        return
    for coord in product(*(range(n) for n in counts)):
        yield tuple(int(v) for v in coord)


def coord_in_bounds(spec: ArraySpec, coord: tuple[int, ...]) -> bool:
    """Check whether a chunk coordinate is valid for this array."""

    counts = chunk_counts(spec)
    if len(coord) != len(counts):
        return False
    return all(0 <= index < count for index, count in zip(coord, counts, strict=True))


def chunk_key(spec: ArraySpec, coord: tuple[int, ...]) -> str:
    """Encode chunk coordinates according to Zarr chunk_key_encoding."""

    if spec.chunk_key_encoding == "default":
        if not coord:
            return "c"
        return f"c{spec.separator}" + spec.separator.join(str(v) for v in coord)
    if spec.chunk_key_encoding == "v2":
        return "0" if not coord else spec.separator.join(str(v) for v in coord)
    raise ValueError(f"Unsupported chunk_key_encoding '{spec.chunk_key_encoding}' for {spec.name}")


def chunk_path(spec: ArraySpec, coord: tuple[int, ...]) -> Path:
    """Return the chunk file path for a coordinate."""

    return spec.path / chunk_key(spec, coord)



#!/usr/bin/env python3
"""Benchmark repeated `xzarrguard check` runs and optionally compare to baseline."""

from __future__ import annotations

import argparse
import json
import os
import statistics
import subprocess
import sys
from datetime import UTC, datetime
from pathlib import Path
from time import perf_counter
from typing import Any

PROJECT_ROOT = Path(__file__).resolve().parents[1]
SRC_PATH = PROJECT_ROOT / "src"


def _summary(values: list[float]) -> dict[str, float]:
    if not values:
        raise ValueError("No benchmark values were collected")
    p95 = values[0]
    if len(values) > 1:
        p95 = statistics.quantiles(values, n=20, method="inclusive")[18]
    payload: dict[str, float] = {
        "min": min(values),
        "max": max(values),
        "mean": statistics.fmean(values),
        "median": statistics.median(values),
        "p95": p95,
    }
    if len(values) > 1:
        payload["stdev"] = statistics.stdev(values)
    return payload


def _run_once(
    store_path: Path,
    *,
    python_bin: str,
    strict_stale: bool,
) -> dict[str, Any]:
    cmd = [
        python_bin,
        "-m",
        "xzarrguard.cli",
        "check",
        str(store_path),
        "--json",
        "--timing",
    ]
    if strict_stale:
        cmd.append("--strict-stale")

    env = os.environ.copy()
    existing = env.get("PYTHONPATH")
    env["PYTHONPATH"] = str(SRC_PATH) if not existing else f"{SRC_PATH}:{existing}"

    wall_start = perf_counter()
    proc = subprocess.run(
        cmd,
        capture_output=True,
        text=True,
        cwd=PROJECT_ROOT,
        env=env,
        check=False,
    )
    wall_s = perf_counter() - wall_start
    if proc.returncode not in (0, 1):
        raise RuntimeError(
            "benchmark run failed with non-check exit code "
            f"{proc.returncode}\nstderr:\n{proc.stderr}\nstdout:\n{proc.stdout}"
        )

    try:
        payload = json.loads(proc.stdout)
    except json.JSONDecodeError as exc:
        raise RuntimeError(
            "failed to parse JSON output from check command\n"
            f"stderr:\n{proc.stderr}\nstdout:\n{proc.stdout}"
        ) from exc

    timing = payload.get("timing")
    if not isinstance(timing, dict):
        raise RuntimeError("check output did not include timing payload")

    return {
        "exit_code": proc.returncode,
        "ok": bool(payload.get("ok", False)),
        "wall_s": wall_s,
        "reported_total_s": float(timing.get("total_s", wall_s)),
        "scan_specs_s": float(timing.get("scan_specs_s", 0.0)),
        "manifest_s": float(timing.get("manifest_s", 0.0)),
        "chunk_scan_s": float(timing.get("chunk_scan_s", 0.0)),
        "exists_calls": int(timing.get("exists_calls", 0)),
    }


def _parse_args(argv: list[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("store_path", type=Path, help="Zarr store path to benchmark")
    parser.add_argument(
        "--runs",
        type=int,
        default=5,
        help="Number of measured runs (default: 5)",
    )
    parser.add_argument(
        "--warmup",
        type=int,
        default=1,
        help="Number of warmup runs (default: 1)",
    )
    parser.add_argument(
        "--python",
        default=sys.executable,
        help="Python interpreter used for benchmark runs (default: current interpreter)",
    )
    parser.add_argument(
        "--strict-stale",
        action="store_true",
        help="Pass --strict-stale to check runs",
    )
    parser.add_argument(
        "--out",
        type=Path,
        help="Optional file path to write detailed benchmark JSON results",
    )
    parser.add_argument(
        "--baseline",
        type=Path,
        help="Optional benchmark JSON file to compare against",
    )
    args = parser.parse_args(argv)
    if args.runs < 1:
        parser.error("--runs must be >= 1")
    if args.warmup < 0:
        parser.error("--warmup must be >= 0")
    return args


def main(argv: list[str] | None = None) -> int:
    args = _parse_args(argv or sys.argv[1:])
    store_path = args.store_path.resolve()
    if not store_path.exists():
        print(f"error: store does not exist: {store_path}", file=sys.stderr)
        return 2

    for index in range(args.warmup):
        _run_once(store_path, python_bin=args.python, strict_stale=args.strict_stale)
        print(f"warmup {index + 1}/{args.warmup} complete")

    runs: list[dict[str, Any]] = []
    for index in range(args.runs):
        result = _run_once(store_path, python_bin=args.python, strict_stale=args.strict_stale)
        runs.append(result)
        print(
            f"run {index + 1}/{args.runs}: "
            f"reported_total={result['reported_total_s']:.3f}s "
            f"wall={result['wall_s']:.3f}s"
        )

    reported = [float(run["reported_total_s"]) for run in runs]
    wall = [float(run["wall_s"]) for run in runs]
    summary = {
        "reported_total_s": _summary(reported),
        "wall_s": _summary(wall),
        "exists_calls": int(runs[0]["exists_calls"]) if runs else 0,
    }

    output: dict[str, Any] = {
        "schema_version": 1,
        "timestamp_utc": datetime.now(UTC).isoformat(),
        "store_path": str(store_path),
        "python": args.python,
        "runs": args.runs,
        "warmup": args.warmup,
        "strict_stale": bool(args.strict_stale),
        "summary": summary,
        "per_run": runs,
    }

    if args.baseline is not None:
        baseline = json.loads(args.baseline.read_text(encoding="utf-8"))
        baseline_mean = float(baseline["summary"]["reported_total_s"]["mean"])
        current_mean = float(summary["reported_total_s"]["mean"])
        delta_pct = ((current_mean - baseline_mean) / baseline_mean) * 100.0
        output["comparison"] = {
            "baseline": str(args.baseline.resolve()),
            "baseline_mean_reported_total_s": baseline_mean,
            "current_mean_reported_total_s": current_mean,
            "delta_mean_pct": delta_pct,
        }

    print(
        "summary: "
        f"mean={summary['reported_total_s']['mean']:.3f}s "
        f"median={summary['reported_total_s']['median']:.3f}s "
        f"p95={summary['reported_total_s']['p95']:.3f}s"
    )
    if "comparison" in output:
        comparison = output["comparison"]
        print(
            "vs baseline: "
            f"{comparison['delta_mean_pct']:+.2f}% "
            f"({comparison['baseline_mean_reported_total_s']:.3f}s -> "
            f"{comparison['current_mean_reported_total_s']:.3f}s)"
        )

    if args.out is not None:
        args.out.parent.mkdir(parents=True, exist_ok=True)
        args.out.write_text(json.dumps(output, indent=2, sort_keys=True) + "\n", encoding="utf-8")
        print(f"wrote benchmark results: {args.out}")

    return 0


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

import inspect
import json
from pathlib import Path

import fsspec
import numpy as np
import pytest
import xarray as xr

from xzarrguard._version import __version__
from xzarrguard.cli import main
from xzarrguard.create import create_store
from xzarrguard.layout import chunk_path, scan_array_specs
from xzarrguard.manifest import dump_no_data_chunks, manifest_path


def _dataset() -> xr.Dataset:
    ds = xr.Dataset(
        {
            "var": (("x", "y"), np.arange(16, dtype=np.float32).reshape(4, 4)),
        },
        coords={"x": np.arange(4), "y": np.arange(4)},
    )
    ds["var"].encoding["chunks"] = (2, 2)
    return ds


def _dataset_with_nan_chunk() -> xr.Dataset:
    ds = _dataset().copy(deep=True)
    ds["var"].values[0:2, 0:2] = np.nan
    return ds


def _write_source_store(dataset: xr.Dataset, store_path: Path) -> None:
    params = inspect.signature(dataset.to_zarr).parameters
    kwargs: dict[str, object] = {"store": store_path, "mode": "w"}
    if "zarr_format" in params:
        kwargs["zarr_format"] = 3
    else:
        kwargs["zarr_version"] = 3
    if "write_empty_chunks" in params:
        kwargs["write_empty_chunks"] = True
    if "consolidated" in params:
        kwargs["consolidated"] = False
    dataset.to_zarr(**kwargs)


def _write_remote_memory_store() -> str:
    fs = fsspec.filesystem("memory")
    store_uri = "memory://cli-remote-store.zarr"
    root = "/cli-remote-store.zarr"
    if fs.exists(root):
        fs.rm(root, recursive=True)

    with fs.open(f"{root}/zarr.json", "w") as handle:
        json.dump(
            {
                "zarr_format": 3,
                "node_type": "group",
                "attributes": {},
                "consolidated_metadata": {
                    "kind": "inline",
                    "must_understand": False,
                    "metadata": {
                        "var": {
                            "zarr_format": 3,
                            "node_type": "array",
                            "shape": [4, 4],
                            "chunk_grid": {
                                "name": "regular",
                                "configuration": {"chunk_shape": [2, 2]},
                            },
                            "chunk_key_encoding": {
                                "name": "default",
                                "configuration": {"separator": "/"},
                            },
                        }
                    },
                },
            },
            handle,
        )
    for coord in ((0, 1), (1, 0), (1, 1)):
        with fs.open(f"{root}/var/c/{coord[0]}/{coord[1]}", "wb") as handle:
            handle.write(b"chunk")
    with fs.open(f"{root}/.xzarrguard/manifests/var.json", "w") as handle:
        json.dump(
            {
                "schema_version": 1,
                "zarr_format": 3,
                "variable": "var",
                "allowed_missing": [{"coord": [0, 0], "key": "c/0/0"}],
            },
            handle,
        )
    return store_uri


def test_cli_check_exit_code_success(tmp_path: Path, capsys) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset(), store, no_data_strategy="empty_chunks")

    code = main(["check", str(store)])
    out = capsys.readouterr().out

    assert code == 0
    assert "PASS" in out


def test_cli_version_option(capsys) -> None:
    with pytest.raises(SystemExit) as exc:
        main(["--version"])

    out = capsys.readouterr().out.strip()
    assert exc.value.code == 0
    assert out == f"xzarrguard {__version__}"


def test_cli_check_exit_code_failure(tmp_path: Path, capsys) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset(), store, no_data_strategy="empty_chunks")

    spec = next(item for item in scan_array_specs(store) if item.name == "var")
    chunk_path(spec, (0, 0)).unlink()

    code = main(["check", str(store)])
    out = capsys.readouterr().out

    assert code == 1
    assert "FAIL" in out


def test_cli_check_json_output(tmp_path: Path, capsys) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset(), store, no_data_strategy="empty_chunks")

    code = main(["check", str(store), "--json"])
    out = capsys.readouterr().out
    payload = json.loads(out)

    assert code == 0
    assert payload["ok"] is True


def test_cli_check_timing_output(tmp_path: Path, capsys) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset(), store, no_data_strategy="empty_chunks")

    code = main(["check", str(store), "--timing"])
    out = capsys.readouterr().out

    assert code == 0
    assert "timing: total=" in out


def test_cli_check_json_output_with_timing(tmp_path: Path, capsys) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset(), store, no_data_strategy="empty_chunks")

    code = main(["check", str(store), "--json", "--timing"])
    out = capsys.readouterr().out
    payload = json.loads(out)

    assert code == 0
    assert payload["ok"] is True
    assert payload["timing"]["exists_calls"] > 0


def test_cli_check_remote_memory_store(capsys) -> None:
    store_uri = _write_remote_memory_store()

    code = main(["check", store_uri])
    out = capsys.readouterr().out

    assert code == 0
    assert "PASS" in out


def test_cli_check_rejects_invalid_storage_option(capsys) -> None:
    code = main(["check", "memory://cli-remote-store.zarr", "--storage-option", "bad-option"])
    err = capsys.readouterr().err

    assert code == 2
    assert "Invalid --storage-option value" in err


def test_cli_create_then_check_roundtrip(tmp_path: Path, capsys) -> None:
    source = tmp_path / "source.zarr"
    target = tmp_path / "target.zarr"
    no_data = tmp_path / "no_data.json"

    _write_source_store(_dataset(), source)
    dump_no_data_chunks(no_data, {"var": [(0, 0)]})

    create_code = main(["create", str(source), str(target), "--no-data", str(no_data)])
    _ = capsys.readouterr()
    check_code = main(["check", str(target)])

    assert create_code == 0
    assert check_code == 0


def test_cli_create_in_place_metadata_then_check(tmp_path: Path, capsys) -> None:
    store = tmp_path / "store.zarr"
    no_data = tmp_path / "no_data.json"
    create_store(_dataset(), store, no_data_strategy="empty_chunks")

    spec = next(item for item in scan_array_specs(store) if item.name == "var")
    chunk_path(spec, (0, 0)).unlink()
    dump_no_data_chunks(no_data, {"var": [(0, 0)]})

    create_code = main([
        "create",
        str(store),
        "--in-place-metadata-only",
        "--no-data",
        str(no_data),
    ])
    create_out = capsys.readouterr().out
    check_code = main(["check", str(store)])

    assert create_code == 0
    assert f"updated: {store}" in create_out
    assert check_code == 0


def test_cli_create_in_place_metadata_rejects_different_target(capsys) -> None:
    create_code = main([
        "create",
        "source.zarr",
        "target.zarr",
        "--in-place-metadata-only",
    ])
    err = capsys.readouterr().err

    assert create_code == 2
    assert "target_store must be omitted or equal to source_zarr" in err


def test_cli_create_in_place_metadata_infers_missing_from_store(tmp_path: Path, capsys) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset(), store, no_data_strategy="empty_chunks")

    spec = next(item for item in scan_array_specs(store) if item.name == "var")
    chunk_path(spec, (0, 0)).unlink()

    create_code = main([
        "create",
        str(store),
        "--in-place-metadata-only",
        "--infer-no-data-from-store",
    ])
    create_out = capsys.readouterr().out
    check_code = main(["check", str(store)])

    assert create_code == 0
    assert f"updated: {store}" in create_out
    assert "manifests: 1" in create_out
    assert check_code == 0


def test_cli_create_in_place_metadata_rejects_infer_with_no_data(tmp_path: Path, capsys) -> None:
    store = tmp_path / "store.zarr"
    no_data = tmp_path / "no_data.json"
    create_store(_dataset(), store, no_data_strategy="empty_chunks")
    dump_no_data_chunks(no_data, {"var": [(0, 0)]})

    create_code = main([
        "create",
        str(store),
        "--in-place-metadata-only",
        "--infer-no-data-from-store",
        "--no-data",
        str(no_data),
    ])
    err = capsys.readouterr().err

    assert create_code == 2
    assert "--infer-no-data-from-store cannot be combined with --no-data" in err


def test_cli_convert_auto_to_manifest(tmp_path: Path, capsys) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset_with_nan_chunk(), store, no_data_strategy="empty_chunks")

    convert_code = main(["convert", str(store)])
    convert_out = capsys.readouterr().out
    check_code = main(["check", str(store)])

    assert convert_code == 0
    assert "direction: materialized_to_manifest" in convert_out
    assert check_code == 0
    assert manifest_path(store, "var").exists()


def test_cli_convert_explicit_manifest_to_materialized(tmp_path: Path, capsys) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset_with_nan_chunk(), store, no_data_strategy="empty_chunks")

    to_manifest_code = main(["convert", str(store)])
    _ = capsys.readouterr()
    to_materialized_code = main([
        "convert",
        str(store),
        "--direction",
        "manifest_to_materialized",
    ])
    to_materialized_out = capsys.readouterr().out
    strict_check_code = main(["check", str(store), "--strict-stale"])

    assert to_manifest_code == 0
    assert to_materialized_code == 0
    assert "direction: manifest_to_materialized" in to_materialized_out
    assert "manifests_removed: 1" in to_materialized_out
    assert strict_check_code == 0
    assert not manifest_path(store, "var").exists()



"""CLI for checking, creating, and converting integrity-aware stores."""

from __future__ import annotations

import argparse
import json
import sys
from collections.abc import Sequence
from pathlib import Path
from typing import Any

import xarray as xr

from ._version import __version__
from .create import convert_store, create_store
from .integrity import check_store
from .manifest import load_no_data_chunks


def _build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="xzarrguard")
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")
    subparsers = parser.add_subparsers(dest="command", required=True)

    check = subparsers.add_parser("check", help="Check store completeness")
    check.add_argument("store_path", help="Path to Zarr store")
    check.add_argument("--json", action="store_true", help="Print JSON report")
    check.add_argument("--timing", action="store_true", help="Print coarse timing details")
    check.add_argument(
        "--strict-stale",
        action="store_true",
        help="Fail when manifest contains entries for chunks that exist",
    )
    check.add_argument(
        "--profile",
        help="Credential profile for remote backends such as s3fs",
    )
    check.add_argument(
        "--endpoint-url",
        help="Custom endpoint URL for S3-compatible backends",
    )
    check.add_argument(
        "--storage-option",
        action="append",
        default=[],
        metavar="KEY=VALUE",
        help=(
            "Additional fsspec storage option. Use dotted keys for nesting, for example "
            "client_kwargs.endpoint_url=https://object-store.example.com"
        ),
    )

    create = subparsers.add_parser("create", help="Create integrity-aware store")
    create.add_argument("source_zarr", help="Source Zarr store readable by xarray")
    create.add_argument(
        "target_store",
        nargs="?",
        help="Target Zarr store path. Omit when using --in-place-metadata-only.",
    )
    create.add_argument("--no-data", help="JSON mapping of variable to no-data chunk coordinates")
    create.add_argument("--overwrite", action="store_true", help="Overwrite target if it exists")
    create.add_argument(
        "--in-place-metadata-only",
        action="store_true",
        help="Update manifests in an existing store without rewriting chunk data",
    )
    create.add_argument(
        "--infer-no-data-from-store",
        action="store_true",
        help="In in-place mode, derive allowed-missing chunks from currently missing chunks",
    )

    convert = subparsers.add_parser("convert", help="Convert store materialization mode")
    convert.add_argument("store_path", help="Path to existing Zarr store")
    convert.add_argument(
        "--direction",
        default="auto",
        choices=["auto", "materialized_to_manifest", "manifest_to_materialized"],
        help=(
            "Conversion direction. auto chooses based on whether manifests exist "
            "(default: %(default)s)"
        ),
    )

    return parser


def _run_check(args: argparse.Namespace) -> int:
    try:
        storage_options = _build_storage_options(args)
        report = check_store(
            args.store_path,
            strict_stale_manifest=args.strict_stale,
            timing=args.timing,
            storage_options=storage_options,
        )
    except Exception as exc:  # pragma: no cover - defensive CLI guard
        print(f"error: {exc}", file=sys.stderr)
        return 2

    if args.json:
        print(json.dumps(report.to_dict(), indent=2, sort_keys=True))
    else:
        print("PASS" if report.ok else "FAIL")
        for name in sorted(report.variables):
            item = report.variables[name]
            details: list[str] = []
            if item.missing_unexpected:
                details.append(f"missing_unexpected={len(item.missing_unexpected)}")
            if item.stale_manifest:
                details.append(f"stale_manifest={len(item.stale_manifest)}")
            if item.manifest_key_mismatch:
                details.append(f"manifest_key_mismatch={len(item.manifest_key_mismatch)}")
            if item.manifest_out_of_bounds:
                details.append(f"manifest_out_of_bounds={len(item.manifest_out_of_bounds)}")
            if details:
                print(f"{name}: {', '.join(details)}")
        for error in report.errors:
            print(f"error: {error}")
        if args.timing and report.timing is not None:
            timing = report.timing
            print(
                "timing: "
                f"total={timing.total_s:.3f}s "
                f"scan_specs={timing.scan_specs_s:.3f}s "
                f"manifest={timing.manifest_s:.3f}s "
                f"chunk_scan={timing.chunk_scan_s:.3f}s "
                f"exists_calls={timing.exists_calls}"
            )

    return 0 if report.ok else 1


def _run_create(args: argparse.Namespace) -> int:
    no_data = load_no_data_chunks(args.no_data) if args.no_data else None

    if args.in_place_metadata_only:
        if args.overwrite:
            print(
                "error: --overwrite cannot be used with --in-place-metadata-only",
                file=sys.stderr,
            )
            return 2
        if args.infer_no_data_from_store and args.no_data:
            print(
                "error: --infer-no-data-from-store cannot be combined with --no-data",
                file=sys.stderr,
            )
            return 2

        if args.target_store is None:
            target_store = args.source_zarr
        else:
            source = Path(args.source_zarr)
            target = Path(args.target_store)
            if source != target:
                print(
                    "error: with --in-place-metadata-only, target_store must be omitted "
                    "or equal to source_zarr",
                    file=sys.stderr,
                )
                return 2
            target_store = args.target_store

        try:
            report = create_store(
                None,
                target_store,
                no_data_chunks=no_data,
                in_place_metadata_only=True,
                infer_no_data_from_store=args.infer_no_data_from_store,
            )
        except Exception as exc:
            print(f"error: {exc}", file=sys.stderr)
            return 2

        print(f"updated: {report.store_path}")
        if report.manifests_written:
            print(f"manifests: {len(report.manifests_written)}")
        elif args.infer_no_data_from_store:
            print("manifests: 0 (no missing chunks detected)")
        return 0

    if args.target_store is None:
        print(
            "error: target_store is required unless --in-place-metadata-only is used",
            file=sys.stderr,
        )
        return 2

    dataset = xr.open_zarr(args.source_zarr, consolidated=False)
    try:
        report = create_store(
            dataset,
            args.target_store,
            no_data_chunks=no_data,
            overwrite=args.overwrite,
        )
    except Exception as exc:
        print(f"error: {exc}", file=sys.stderr)
        return 2
    finally:
        dataset.close()

    print(f"created: {report.store_path}")
    if report.manifests_written:
        print(f"manifests: {len(report.manifests_written)}")
    return 0


def _run_convert(args: argparse.Namespace) -> int:
    try:
        report = convert_store(args.store_path, direction=args.direction)
    except Exception as exc:
        print(f"error: {exc}", file=sys.stderr)
        return 2

    print(f"converted: {report.store_path}")
    print(f"direction: {report.direction}")
    if report.manifests_written:
        print(f"manifests_written: {len(report.manifests_written)}")
    if report.manifests_removed:
        print(f"manifests_removed: {len(report.manifests_removed)}")
    deleted_count = sum(len(refs) for refs in report.deleted_chunks.values())
    materialized_count = sum(len(refs) for refs in report.materialized_chunks.values())
    if deleted_count:
        print(f"deleted_chunks: {deleted_count}")
    if materialized_count:
        print(f"materialized_chunks: {materialized_count}")
    return 0


def _build_storage_options(args: argparse.Namespace) -> dict[str, Any]:
    options = _parse_storage_options(args.storage_option)
    if args.profile:
        options.setdefault("profile", args.profile)
    if args.endpoint_url:
        client_kwargs = options.setdefault("client_kwargs", {})
        if not isinstance(client_kwargs, dict):
            raise ValueError("client_kwargs storage option must be a mapping")
        client_kwargs.setdefault("endpoint_url", args.endpoint_url)
    return options


def _parse_storage_options(items: Sequence[str]) -> dict[str, Any]:
    options: dict[str, Any] = {}
    for item in items:
        key, separator, raw_value = item.partition("=")
        if not separator or not key:
            raise ValueError(f"Invalid --storage-option value: {item!r}")
        _assign_nested_option(options, key.split("."), _coerce_storage_value(raw_value))
    return options


def _assign_nested_option(target: dict[str, Any], path: Sequence[str], value: Any) -> None:
    current = target
    for key in path[:-1]:
        existing = current.setdefault(key, {})
        if not isinstance(existing, dict):
            raise ValueError(f"Storage option path conflict at {'.'.join(path)}")
        current = existing
    current[path[-1]] = value


def _coerce_storage_value(value: str) -> Any:
    lowered = value.lower()
    if lowered in {"true", "false", "null"}:
        return json.loads(lowered)
    try:
        if value and value[0] in {"[", "{", "\""}:
            return json.loads(value)
        if value.isdigit() or (value.startswith("-") and value[1:].isdigit()):
            return json.loads(value)
        if "." in value:
            return json.loads(value)
    except json.JSONDecodeError:
        pass
    return value


def main(argv: Sequence[str] | None = None) -> int:
    """CLI entrypoint."""

    parser = _build_parser()
    args = parser.parse_args(argv)

    if args.command == "check":
        return _run_check(args)
    if args.command == "create":
        return _run_create(args)
    if args.command == "convert":
        return _run_convert(args)
    parser.error(f"unknown command: {args.command}")
    return 2


if __name__ == "__main__":  # pragma: no cover
    raise SystemExit(main())



"""Store completeness checks for local and remote Zarr v3 stores."""

from __future__ import annotations

from collections.abc import Mapping
from pathlib import Path
from time import perf_counter
from typing import Any

from .layout import chunk_key, chunk_path, coord_in_bounds, expected_chunk_coords, scan_array_specs
from .manifest import MANIFEST_ROOT, load_variable_manifest
from .models import ChunkRef, IntegrityReport, IntegrityTiming, VariableIntegrity, VariableTiming
from .remote import (
    is_remote_store_path,
    remote_existing_chunk_keys,
    load_remote_variable_manifest,
    open_remote_store,
    scan_remote_array_specs,
)


def check_store(
    store_path: str | Path,
    *,
    strict_stale_manifest: bool = False,
    timing: bool = False,
    storage_options: Mapping[str, Any] | None = None,
    _manifest_root: str | Path = MANIFEST_ROOT,
) -> IntegrityReport:
    """Validate completeness of a Zarr v3 store."""

    if is_remote_store_path(store_path):
        return _check_remote_store(
            str(store_path),
            strict_stale_manifest=strict_stale_manifest,
            timing=timing,
            storage_options=storage_options,
            _manifest_root=_manifest_root,
        )

    total_start = perf_counter() if timing else 0.0
    store = Path(store_path)
    report = IntegrityReport(store_path=str(store), strict_stale_manifest=strict_stale_manifest)
    timing_data = IntegrityTiming() if timing else None

    def _finish() -> IntegrityReport:
        if timing_data is not None:
            timing_data.total_s = perf_counter() - total_start
            report.timing = timing_data
        return report

    if not store.exists():
        report.ok = False
        report.errors.append(f"Store does not exist: {store}")
        return _finish()
    if not store.is_dir():
        report.ok = False
        report.errors.append(f"Store path is not a directory: {store}")
        return _finish()

    scan_start = perf_counter() if timing_data is not None else 0.0
    try:
        specs = scan_array_specs(store)
    except ValueError as exc:
        report.ok = False
        report.errors.append(str(exc))
        return _finish()
    if timing_data is not None:
        timing_data.scan_specs_s = perf_counter() - scan_start

    for spec in specs:
        var_timing = VariableTiming() if timing_data is not None else None

        manifest_load_start = perf_counter() if var_timing is not None else 0.0
        has_manifest, manifest_refs = load_variable_manifest(
            store,
            spec.name,
            manifest_root=_manifest_root,
        )
        if var_timing is not None:
            var_timing.manifest_load_s = perf_counter() - manifest_load_start

        variable = VariableIntegrity(
            name=spec.name,
            expected_chunks=0,
            has_manifest=has_manifest,
        )

        manifest_validate_start = perf_counter() if var_timing is not None else 0.0
        valid_manifest: dict[tuple[int, ...], ChunkRef] = {}
        for ref in manifest_refs:
            if not coord_in_bounds(spec, ref.coord):
                variable.manifest_out_of_bounds.append(ref)
                continue
            expected_key = chunk_key(spec, ref.coord)
            if ref.key != expected_key:
                variable.manifest_key_mismatch.append(ref)
                continue
            valid_manifest[ref.coord] = ref

        if var_timing is not None:
            var_timing.manifest_validate_s = perf_counter() - manifest_validate_start

        chunk_scan_start = perf_counter() if var_timing is not None else 0.0
        for coord in expected_chunk_coords(spec):
            variable.expected_chunks += 1
            key = chunk_key(spec, coord)
            exists = chunk_path(spec, coord).exists()
            ref = ChunkRef(coord=coord, key=key)

            if not exists and coord in valid_manifest:
                variable.missing_allowed.append(ref)
            elif not exists:
                variable.missing_unexpected.append(ref)
            elif coord in valid_manifest:
                variable.stale_manifest.append(ref)

        if var_timing is not None:
            var_timing.chunk_scan_s = perf_counter() - chunk_scan_start
            var_timing.expected_chunks = variable.expected_chunks
            var_timing.missing_chunks = len(variable.missing_allowed) + len(
                variable.missing_unexpected
            )

        variable.ok = not variable.missing_unexpected
        if variable.manifest_key_mismatch or variable.manifest_out_of_bounds:
            variable.ok = False
        if strict_stale_manifest and variable.stale_manifest:
            variable.ok = False

        report.variables[spec.name] = variable
        if timing_data is not None and var_timing is not None:
            timing_data.variables[spec.name] = var_timing
            timing_data.manifest_s += var_timing.manifest_load_s + var_timing.manifest_validate_s
            timing_data.chunk_scan_s += var_timing.chunk_scan_s
            timing_data.exists_calls += variable.expected_chunks

    report.ok = not report.errors and all(variable.ok for variable in report.variables.values())
    return _finish()


def _check_remote_store(
    store_path: str,
    *,
    strict_stale_manifest: bool = False,
    timing: bool = False,
    storage_options: Mapping[str, Any] | None = None,
    _manifest_root: str | Path = MANIFEST_ROOT,
) -> IntegrityReport:
    total_start = perf_counter() if timing else 0.0
    report = IntegrityReport(store_path=store_path, strict_stale_manifest=strict_stale_manifest)
    timing_data = IntegrityTiming() if timing else None

    def _finish() -> IntegrityReport:
        if timing_data is not None:
            timing_data.total_s = perf_counter() - total_start
            report.timing = timing_data
        return report

    try:
        fs, root_key = open_remote_store(store_path, storage_options=storage_options)
    except Exception as exc:
        report.ok = False
        report.errors.append(str(exc))
        return _finish()

    root_meta_key = f"{root_key.rstrip('/')}/zarr.json" if root_key not in {"", "/"} else "zarr.json"
    if not fs.exists(root_key) and not fs.exists(root_meta_key):
        report.ok = False
        report.errors.append(f"Store does not exist: {store_path}")
        return _finish()

    scan_start = perf_counter() if timing_data is not None else 0.0
    try:
        specs = scan_remote_array_specs(fs, root_key)
    except ValueError as exc:
        report.ok = False
        report.errors.append(str(exc))
        return _finish()
    if timing_data is not None:
        timing_data.scan_specs_s = perf_counter() - scan_start

    for spec in specs:
        var_timing = VariableTiming() if timing_data is not None else None

        manifest_load_start = perf_counter() if var_timing is not None else 0.0
        has_manifest, manifest_refs = load_remote_variable_manifest(
            fs,
            root_key,
            spec.name,
            manifest_root=_manifest_root,
        )
        if var_timing is not None:
            var_timing.manifest_load_s = perf_counter() - manifest_load_start

        variable = VariableIntegrity(
            name=spec.name,
            expected_chunks=0,
            has_manifest=has_manifest,
        )

        manifest_validate_start = perf_counter() if var_timing is not None else 0.0
        valid_manifest: dict[tuple[int, ...], ChunkRef] = {}
        for ref in manifest_refs:
            if not coord_in_bounds(spec, ref.coord):
                variable.manifest_out_of_bounds.append(ref)
                continue
            expected_key = chunk_key(spec, ref.coord)
            if ref.key != expected_key:
                variable.manifest_key_mismatch.append(ref)
                continue
            valid_manifest[ref.coord] = ref

        if var_timing is not None:
            var_timing.manifest_validate_s = perf_counter() - manifest_validate_start

        chunk_scan_start = perf_counter() if var_timing is not None else 0.0
        existing_chunk_keys = remote_existing_chunk_keys(fs, spec)
        for coord in expected_chunk_coords(spec):
            variable.expected_chunks += 1
            key = chunk_key(spec, coord)
            exists = key in existing_chunk_keys
            ref = ChunkRef(coord=coord, key=key)

            if not exists and coord in valid_manifest:
                variable.missing_allowed.append(ref)
            elif not exists:
                variable.missing_unexpected.append(ref)
            elif coord in valid_manifest:
                variable.stale_manifest.append(ref)

        if var_timing is not None:
            var_timing.chunk_scan_s = perf_counter() - chunk_scan_start
            var_timing.expected_chunks = variable.expected_chunks
            var_timing.missing_chunks = len(variable.missing_allowed) + len(
                variable.missing_unexpected
            )

        variable.ok = not variable.missing_unexpected
        if variable.manifest_key_mismatch or variable.manifest_out_of_bounds:
            variable.ok = False
        if strict_stale_manifest and variable.stale_manifest:
            variable.ok = False

        report.variables[spec.name] = variable
        if timing_data is not None and var_timing is not None:
            timing_data.variables[spec.name] = var_timing
            timing_data.manifest_s += var_timing.manifest_load_s + var_timing.manifest_validate_s
            timing_data.chunk_scan_s += var_timing.chunk_scan_s
            timing_data.exists_calls += variable.expected_chunks

    report.ok = not report.errors and all(variable.ok for variable in report.variables.values())
    return _finish()



from __future__ import annotations

import inspect
import json
from pathlib import Path

import fsspec
import numpy as np
import pytest
import xarray as xr

from xzarrguard import (
    ConvertDirection,
    ConvertReport,
    NoDataStrategy,
    check_store,
    convert_store,
    create_store,
    guarded_to_zarr,
)
from xzarrguard.layout import chunk_key, chunk_path, scan_array_specs
from xzarrguard.manifest import (
    load_no_data_chunks,
    load_variable_manifest,
    manifest_path,
    write_variable_manifest,
)
from xzarrguard.models import (
    ChunkRef,
)
from xzarrguard.models import (
    ConvertDirection as ModelsConvertDirection,
)
from xzarrguard.models import (
    NoDataStrategy as ModelsNoDataStrategy,
)


def _dataset() -> xr.Dataset:
    ds = xr.Dataset(
        {
            "var": (("x", "y"), np.arange(16, dtype=np.float32).reshape(4, 4)),
        },
        coords={"x": np.arange(4), "y": np.arange(4)},
    )
    ds["var"].encoding["chunks"] = (2, 2)
    return ds


def _dataset_with_nan_chunk() -> xr.Dataset:
    ds = _dataset().copy(deep=True)
    ds["var"].values[0:2, 0:2] = np.nan
    return ds


def _delete_chunk(store_path: Path, variable: str, coord: tuple[int, ...]) -> None:
    spec = next(item for item in scan_array_specs(store_path) if item.name == variable)
    path = chunk_path(spec, coord)
    path.unlink()


def _write_source_store(dataset: xr.Dataset, store_path: Path) -> None:
    params = inspect.signature(dataset.to_zarr).parameters
    kwargs: dict[str, object] = {"store": store_path, "mode": "w"}
    if "zarr_format" in params:
        kwargs["zarr_format"] = 3
    else:
        kwargs["zarr_version"] = 3
    if "write_empty_chunks" in params:
        kwargs["write_empty_chunks"] = True
    if "consolidated" in params:
        kwargs["consolidated"] = False
    dataset.to_zarr(**kwargs)


def _remote_group_meta(*, metadata: dict[str, dict] | None = None) -> dict:
    payload: dict[str, object] = {"zarr_format": 3, "node_type": "group", "attributes": {}}
    if metadata is not None:
        payload["consolidated_metadata"] = {
            "kind": "inline",
            "must_understand": False,
            "metadata": metadata,
        }
    return payload


def _remote_array_meta(shape: tuple[int, ...], chunk_shape: tuple[int, ...]) -> dict:
    return {
        "zarr_format": 3,
        "node_type": "array",
        "shape": list(shape),
        "chunk_grid": {"name": "regular", "configuration": {"chunk_shape": list(chunk_shape)}},
        "chunk_key_encoding": {"name": "default", "configuration": {"separator": "/"}},
    }


def _write_remote_memory_store(
    store_uri: str,
    *,
    missing_manifested: list[tuple[int, ...]] | None = None,
    missing_unexpected: list[tuple[int, ...]] | None = None,
) -> None:
    fs = fsspec.filesystem("memory")
    root = "/" + store_uri.split("://", 1)[1].lstrip("/")
    if fs.exists(root):
        fs.rm(root, recursive=True)

    with fs.open(f"{root}/zarr.json", "w") as handle:
        json.dump(
            _remote_group_meta(metadata={"var": _remote_array_meta((4, 4), (2, 2))}),
            handle,
        )

    missing_manifested = missing_manifested or []
    missing_unexpected = missing_unexpected or []
    for coord in ((0, 0), (0, 1), (1, 0), (1, 1)):
        if coord in missing_manifested or coord in missing_unexpected:
            continue
        with fs.open(f"{root}/var/c/{coord[0]}/{coord[1]}", "wb") as handle:
            handle.write(b"chunk")

    if missing_manifested:
        with fs.open(f"{root}/.xzarrguard/manifests/var.json", "w") as handle:
            json.dump(
                {
                    "schema_version": 1,
                    "zarr_format": 3,
                    "variable": "var",
                    "allowed_missing": [
                        {"coord": list(coord), "key": f"c/{coord[0]}/{coord[1]}"}
                        for coord in missing_manifested
                    ],
                },
                handle,
            )


def test_check_passes_complete_store_without_manifest(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset(), store, no_data_strategy="empty_chunks")

    report = check_store(store)

    assert report.ok
    assert bool(report)


def test_public_api_exports_convert_models() -> None:
    assert ConvertDirection is ModelsConvertDirection
    assert NoDataStrategy is ModelsNoDataStrategy
    assert ConvertReport(store_path="store.zarr", direction="materialized_to_manifest").ok


def test_check_with_timing_populates_timing_payload(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset(), store, no_data_strategy="empty_chunks")

    report = check_store(store, timing=True)

    assert report.timing is not None
    assert report.timing.total_s >= 0.0
    expected_exists_calls = sum(item.expected_chunks for item in report.variables.values())
    assert report.timing.exists_calls == expected_exists_calls
    assert "var" in report.timing.variables
    var_timing = report.timing.variables["var"]
    assert var_timing.expected_chunks == report.variables["var"].expected_chunks
    assert var_timing.chunk_scan_s >= 0.0
    payload = report.to_dict()
    assert payload["timing"]["exists_calls"] == expected_exists_calls


def test_check_without_timing_omits_timing_payload(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset(), store, no_data_strategy="empty_chunks")

    report = check_store(store)

    assert report.timing is None
    assert "timing" not in report.to_dict()


def test_check_fails_missing_chunk_without_manifest(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset(), store, no_data_strategy="empty_chunks")
    _delete_chunk(store, "var", (0, 0))

    report = check_store(store)

    assert not report.ok
    assert not bool(report)
    assert report.variables["var"].missing_unexpected


def test_check_passes_when_missing_is_manifested(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset(), store, no_data_chunks={"var": [(0, 0)]})

    report = check_store(store)

    assert report.ok
    assert len(report.variables["var"].missing_allowed) == 1


def test_check_remote_memory_store_passes_when_missing_is_manifested() -> None:
    store_uri = "memory://remote-guarded.zarr"
    _write_remote_memory_store(store_uri, missing_manifested=[(0, 0)])

    report = check_store(store_uri)

    assert report.ok
    assert len(report.variables["var"].missing_allowed) == 1


def test_check_remote_memory_store_fails_when_missing_is_unexpected() -> None:
    store_uri = "memory://remote-guarded.zarr"
    _write_remote_memory_store(store_uri, missing_unexpected=[(1, 1)])

    report = check_store(store_uri)

    assert not report.ok
    assert any(item.coord == (1, 1) for item in report.variables["var"].missing_unexpected)


def test_check_fails_when_missing_not_manifested(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset(), store, no_data_chunks={"var": [(0, 0)]})
    _delete_chunk(store, "var", (1, 1))

    report = check_store(store)

    assert not report.ok
    assert any(item.coord == (1, 1) for item in report.variables["var"].missing_unexpected)


def test_stale_manifest_behavior_strict_and_non_strict(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset(), store, no_data_strategy="empty_chunks")

    spec = next(item for item in scan_array_specs(store) if item.name == "var")
    write_variable_manifest(
        store,
        "var",
        [ChunkRef(coord=(0, 0), key=chunk_key(spec, (0, 0)))],
    )

    loose = check_store(store)
    strict = check_store(store, strict_stale_manifest=True)

    assert loose.ok
    assert loose.variables["var"].stale_manifest
    assert not strict.ok


def test_create_manifest_strategy_roundtrip(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"

    report = create_store(
        _dataset(),
        store,
        no_data_chunks={"var": [(0, 1)]},
        no_data_strategy="manifest",
    )
    check = check_store(store)

    assert report.ok
    assert report.manifests_written
    assert check.ok


def test_create_empty_chunks_strategy_roundtrip(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"

    report = create_store(
        _dataset(),
        store,
        no_data_chunks={"var": [(0, 1)]},
        no_data_strategy="empty_chunks",
    )
    check = check_store(store)

    assert report.ok
    assert not report.manifests_written
    assert check.ok


def test_guarded_to_zarr_roundtrip(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"

    report = guarded_to_zarr(_dataset(), store)
    check = check_store(store)

    assert report.ok
    assert check.ok


def test_guarded_to_zarr_rejects_store_kwarg(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"

    with pytest.raises(ValueError, match="must not be provided"):
        guarded_to_zarr(
            _dataset(),
            store,
            to_zarr_kwargs={"store": store},
        )


def test_convert_materialized_to_manifest_and_back(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset_with_nan_chunk(), store, no_data_strategy="empty_chunks")

    spec = next(item for item in scan_array_specs(store) if item.name == "var")
    nan_coord = (0, 0)
    assert chunk_path(spec, nan_coord).exists()

    to_manifest = convert_store(store, direction="materialized_to_manifest")
    assert to_manifest.ok
    assert to_manifest.direction == "materialized_to_manifest"
    assert any(ref.coord == nan_coord for ref in to_manifest.deleted_chunks.get("var", []))
    assert manifest_path(store, "var").exists()
    assert not chunk_path(spec, nan_coord).exists()
    assert check_store(store).ok

    to_materialized = convert_store(store, direction="manifest_to_materialized")
    assert to_materialized.ok
    assert to_materialized.direction == "manifest_to_materialized"
    assert to_materialized.manifests_removed
    assert any(
        ref.coord == nan_coord
        for ref in to_materialized.materialized_chunks.get("var", [])
    )
    assert not manifest_path(store, "var").exists()
    assert chunk_path(spec, nan_coord).exists()
    assert check_store(store, strict_stale_manifest=True).ok

    reopened = xr.open_zarr(store, consolidated=False)
    try:
        assert np.isnan(reopened["var"].isel(x=slice(0, 2), y=slice(0, 2)).values).all()
    finally:
        reopened.close()


def test_convert_auto_direction_switches_on_manifest_presence(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset_with_nan_chunk(), store, no_data_strategy="empty_chunks")

    first = convert_store(store)
    second = convert_store(store)

    assert first.direction == "materialized_to_manifest"
    assert second.direction == "manifest_to_materialized"
    assert check_store(store, strict_stale_manifest=True).ok


def test_create_in_place_metadata_roundtrip(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset(), store, no_data_strategy="empty_chunks")

    spec = next(item for item in scan_array_specs(store) if item.name == "var")
    missing_coord = (0, 0)
    chunk_path(spec, missing_coord).unlink()

    report = create_store(
        None,
        store,
        no_data_chunks={"var": [missing_coord]},
        in_place_metadata_only=True,
    )

    assert report.ok
    assert report.manifests_written
    assert not chunk_path(spec, missing_coord).exists()
    assert check_store(store).ok


def test_create_in_place_metadata_requires_missing_chunks(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset(), store, no_data_strategy="empty_chunks")

    with pytest.raises(ValueError, match="currently missing chunks"):
        create_store(
            None,
            store,
            no_data_chunks={"var": [(0, 0)]},
            in_place_metadata_only=True,
        )


def test_create_in_place_metadata_is_transactional(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset(), store, no_data_chunks={"var": [(0, 1)]}, no_data_strategy="manifest")

    manifest_file = manifest_path(store, "var")
    before = manifest_file.read_text(encoding="utf-8")

    with pytest.raises(ValueError, match="out of bounds"):
        create_store(
            None,
            store,
            no_data_chunks={"var": [(99, 99)]},
            in_place_metadata_only=True,
        )

    after = manifest_file.read_text(encoding="utf-8")
    assert before == after


def test_create_in_place_metadata_infers_missing_from_store(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset(), store, no_data_strategy="empty_chunks")

    spec = next(item for item in scan_array_specs(store) if item.name == "var")
    chunk_path(spec, (0, 0)).unlink()
    chunk_path(spec, (1, 1)).unlink()

    report = create_store(
        None,
        store,
        in_place_metadata_only=True,
        infer_no_data_from_store=True,
    )

    assert report.ok
    assert report.manifests_written
    has_manifest, refs = load_variable_manifest(store, "var")
    assert has_manifest
    assert sorted(ref.coord for ref in refs) == [(0, 0), (1, 1)]
    assert check_store(store).ok


def test_create_in_place_metadata_infer_replaces_stale_manifests(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset(), store, no_data_strategy="empty_chunks")

    spec = next(item for item in scan_array_specs(store) if item.name == "var")
    write_variable_manifest(
        store,
        "var",
        [ChunkRef(coord=(0, 0), key=chunk_key(spec, (0, 0)))],
    )
    assert manifest_path(store, "var").exists()

    report = create_store(
        None,
        store,
        in_place_metadata_only=True,
        infer_no_data_from_store=True,
    )

    assert report.ok
    assert not report.manifests_written
    assert not manifest_path(store, "var").exists()
    assert check_store(store, strict_stale_manifest=True).ok


def test_create_in_place_metadata_infer_rejects_explicit_no_data(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"
    create_store(_dataset(), store, no_data_strategy="empty_chunks")

    with pytest.raises(ValueError, match="cannot be combined with explicit no_data_chunks"):
        create_store(
            None,
            store,
            no_data_chunks={"var": [(0, 0)]},
            in_place_metadata_only=True,
            infer_no_data_from_store=True,
        )


def test_load_no_data_mapping_validation(tmp_path: Path) -> None:
    path = tmp_path / "bad.json"
    path.write_text("[]", encoding="utf-8")

    with pytest.raises(ValueError):
        load_no_data_chunks(path)


def test_create_unknown_variable_fails(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"

    with pytest.raises(ValueError):
        create_store(_dataset(), store, no_data_chunks={"missing": [(0, 0)]})


def test_helper_write_source_store(tmp_path: Path) -> None:
    source = tmp_path / "source.zarr"

    _write_source_store(_dataset(), source)

    assert source.exists()


def test_check_consolidated_metadata_without_child_metadata_files(tmp_path: Path) -> None:
    store = tmp_path / "store.zarr"
    (store / "var" / "c").mkdir(parents=True)
    root_payload = {
        "zarr_format": 3,
        "node_type": "group",
        "attributes": {},
        "consolidated_metadata": {
            "kind": "inline",
            "must_understand": False,
            "metadata": {
                "var": {
                    "zarr_format": 3,
                    "node_type": "array",
                    "shape": [2],
                    "chunk_grid": {"name": "regular", "configuration": {"chunk_shape": [1]}},
                    "chunk_key_encoding": {
                        "name": "default",
                        "configuration": {"separator": "/"},
                    },
                    "codecs": [],
                    "data_type": "int32",
                    "fill_value": 0,
                    "attributes": {},
                    "dimension_names": ["x"],
                    "storage_transformers": [],
                }
            },
        },
    }
    (store / "zarr.json").write_text(json.dumps(root_payload), encoding="utf-8")
    (store / "var" / "c" / "0").write_bytes(b"0")
    (store / "var" / "c" / "1").write_bytes(b"1")

    report_ok = check_store(store)
    assert report_ok.ok

    (store / "var" / "c" / "1").unlink()
    report_fail = check_store(store)
    assert not report_fail.ok
    assert len(report_fail.variables["var"].missing_unexpected) == 1



"""Manifest read/write helpers."""

from __future__ import annotations

import json
from collections.abc import Iterable, Mapping
from pathlib import Path
from urllib.parse import quote

from .models import ChunkRef

MANIFEST_SCHEMA_VERSION = 1
MANIFEST_ROOT = Path(".xzarrguard") / "manifests"


def _normalize_coord(coord: Iterable[int]) -> tuple[int, ...]:
    return tuple(int(item) for item in coord)


def _normalize_mapping(
    mapping: Mapping[str, Iterable[Iterable[int]]] | None,
) -> dict[str, list[tuple[int, ...]]]:
    normalized: dict[str, list[tuple[int, ...]]] = {}
    if not mapping:
        return normalized

    for variable, coords in mapping.items():
        parsed = [_normalize_coord(coord) for coord in coords]
        unique = sorted(set(parsed))
        normalized[str(variable)] = unique
    return normalized


def load_no_data_chunks(path: str | Path) -> dict[str, list[tuple[int, ...]]]:
    """Load variable->chunk-coordinate mapping from JSON."""

    payload = json.loads(Path(path).read_text(encoding="utf-8"))
    if not isinstance(payload, dict):
        raise ValueError("No-data mapping must be an object")
    mapping = {
        str(name): [tuple(int(v) for v in coord) for coord in coords]
        for name, coords in payload.items()
    }
    return _normalize_mapping(mapping)


def dump_no_data_chunks(
    path: str | Path,
    mapping: Mapping[str, Iterable[Iterable[int]]] | None,
) -> None:
    """Write variable->chunk-coordinate mapping to JSON."""

    normalized = _normalize_mapping(mapping)
    serializable = {
        variable: [list(coord) for coord in coords]
        for variable, coords in sorted(normalized.items(), key=lambda item: item[0])
    }
    Path(path).write_text(
        json.dumps(serializable, indent=2, sort_keys=True) + "\n",
        encoding="utf-8",
    )


def manifest_path(
    store_path: str | Path,
    variable: str,
    *,
    manifest_root: str | Path = MANIFEST_ROOT,
) -> Path:
    """Return manifest path for one variable."""

    safe_name = quote(variable, safe="")
    return Path(store_path) / Path(manifest_root) / f"{safe_name}.json"


def load_variable_manifest(
    store_path: str | Path,
    variable: str,
    *,
    manifest_root: str | Path = MANIFEST_ROOT,
) -> tuple[bool, list[ChunkRef]]:
    """Read a single variable manifest."""

    path = manifest_path(store_path, variable, manifest_root=manifest_root)
    if not path.exists():
        return False, []

    payload = json.loads(path.read_text(encoding="utf-8"))
    if payload.get("schema_version") != MANIFEST_SCHEMA_VERSION:
        raise ValueError(f"Unsupported manifest schema in {path}")

    items = payload.get("allowed_missing", [])
    refs = [ChunkRef(coord=_normalize_coord(item["coord"]), key=str(item["key"])) for item in items]
    return True, refs


def write_variable_manifest(
    store_path: str | Path,
    variable: str,
    refs: Iterable[ChunkRef],
    *,
    zarr_format: int = 3,
    manifest_root: str | Path = MANIFEST_ROOT,
) -> Path:
    """Write a single variable manifest file."""

    path = manifest_path(store_path, variable, manifest_root=manifest_root)
    path.parent.mkdir(parents=True, exist_ok=True)
    payload = {
        "schema_version": MANIFEST_SCHEMA_VERSION,
        "zarr_format": zarr_format,
        "variable": variable,
        "allowed_missing": [{"coord": list(ref.coord), "key": ref.key} for ref in refs],
    }
    path.write_text(json.dumps(payload, indent=2, sort_keys=True) + "\n", encoding="utf-8")
    return path
